from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator, spinner
from typing import Any, Dict
from ibm_ray_config.modules.utils import get_option_from_list, find_default, find_obj


class ProfileConfig(ConfigBuilder):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)

    @update_decorator
    def run(self) -> Dict[str, Any]:

        @spinner
        def get_instance_profile_objects():
            return self.ibm_vpc_client.list_instance_profiles().get_result()['profiles']

        instance_profile_objects = get_instance_profile_objects()

        default = find_default(
            self.base_config, instance_profile_objects, name='instance_profile_name')
        instance_profile = get_option_from_list(
            'Carefully choose instance profile, please refer to https://cloud.ibm.com/docs/vpc?topic=vpc-profiles', instance_profile_objects, default=default)

        return instance_profile['name']
    
    @update_decorator
    def verify(self, base_config):
        profile_name = self.defaults['profile_name']
        instance_profile_objects = self.ibm_vpc_client.list_instance_profiles().get_result()['profiles']
        profile = find_obj(instance_profile_objects, 'dummy', obj_name=profile_name)
        if not profile:
            raise Exception(f'Specified profile {profile_name} not found in the profile list {instance_profile_objects}')
        return profile_name

    @update_decorator    
    def create_default(self):
        return 'bx2-2x8'



from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator, spinner
from typing import Any, Dict
from ibm_ray_config.modules.utils import get_option_from_list


class EndpointConfig(ConfigBuilder):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        
    @spinner
    def _get_regions_objects(self):
        return ConfigBuilder.ibm_vpc_client.list_regions().get_result()['regions']
        
    @update_decorator
    def run(self) -> Dict[str, Any]:

        regions_objects = self._get_regions_objects()
        
        default = self.defaults.get('region')
        region_obj = get_option_from_list("Choose region", regions_objects, default = default)

        # update global ibm_vpc_client to selected endpoint
        ConfigBuilder.ibm_vpc_client.set_service_url(region_obj['endpoint'] + '/v1')
        ConfigBuilder.region = region_obj['name']
        
        return region_obj['endpoint']
    
    @update_decorator
    def create_default(self):
        # update global ibm_vpc_client to selected endpoint
        regions_objects = self._get_regions_objects()
        
        # currently hardcoded for us-south
        region_obj = next((r for r in regions_objects if r['name'] == 'us-south'), None)
        
        ConfigBuilder.ibm_vpc_client.set_service_url(region_obj['endpoint'] + '/v1')
        ConfigBuilder.region = region_obj['name']
        
        return region_obj['endpoint']



from typing import Any, Dict
from ibm_ray_config.modules.gen2.profile import ProfileConfig

class RayProfileConfig(ProfileConfig):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.defaults['profile_name'] = self.base_config['available_node_types'][available_node_type][
                    'node_config'].get('instance_profile_name')
                break
    
    def update_config(self, profile_name):

        # cpu number based on profile
        cpu_num = int(profile_name.split('-')[1].split('x')[0])

        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.base_config['available_node_types'][available_node_type][
                    'node_config']['instance_profile_name'] = profile_name
                self.base_config['available_node_types'][available_node_type]['resources']['CPU'] = cpu_num
        else:
            self.base_config['available_node_types']['ray_head_default']['node_config']['instance_profile_name'] = profile_name
            self.base_config['available_node_types']['ray_head_default']['resources']['CPU'] = cpu_num




import os
import subprocess
from typing import Any, Dict
from pathlib import Path

import inquirer
from inquirer import errors
from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator, spinner
from ibm_ray_config.modules.utils import (Color, color_msg, find_default, find_name_id,
                                        validate_exists, validate_not_empty)

from ibm_cloud_sdk_core import ApiException
DEFAULT_KEY_NAME = 'default-ssh-key'
        
def generate_keypair(keyname):
    """Returns newly generated public ssh-key's contents and private key's path"""
    home = str(Path.home())
    filename = f"{home}{os.sep}.ssh{os.sep}id.rsa.{keyname}"
    try:
        os.remove(filename)
    except Exception:
        pass

    os.system(f'ssh-keygen -b 2048 -t rsa -f {filename} -q -N ""')
    print(f"\n\n\033[92mSSH key pair been generated\n")
    print(f"private key: {os.path.abspath(filename)}")
    print(f"public key {os.path.abspath(filename)}.pub\033[0m")
    with open(f"{filename}.pub", 'r') as file:
        ssh_key_data = file.read()
    ssh_key_path = os.path.abspath(filename)
    return ssh_key_data, ssh_key_path

def get_ssh_key(ibm_vpc_client, name):
    """Returns ssh key matching specified name, stored in the VPC associated with the vpc_client"""
    for key in ibm_vpc_client.list_keys().result['keys']:
        if key['name'] == name:
            return key
                    
def register_ssh_key(ibm_vpc_client, config, auto=False):
    """Returns the key's name on the VPC platform, it's public key's contents and the local path to it.
        Registers an either existing or newly generated ssh-key to a specific VPC. """
    if config.get('ibm_vpc'):
        resource_group_id = config['ibm_vpc']['resource_group_id']
    else:
        for available_node_type in config['available_node_types']:
            resource_group_id = config['available_node_types'][available_node_type]['node_config']['resource_group_id']
            break

    
    questions = [
            inquirer.Text(
                'keyname', message='Please specify a name for the new key', validate=validate_not_empty)
    ]
    
    answers = {}
    if not auto:
        answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
    else:
        answers['keyname'] = DEFAULT_KEY_NAME

    keyname = answers['keyname']

    EXISTING_CONTENTS = 'Paste existing public key contents'
    EXISTING_PATH = 'Provide path to existing public key'
    GENERATE_NEW = 'Generate new public key'

    questions = [
        inquirer.List('answer',
                      message="Please choose",
                      choices=[EXISTING_PATH, EXISTING_CONTENTS, GENERATE_NEW]
                      )]

    if not auto:
        answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
    else:
        answers["answer"] = GENERATE_NEW

    ssh_key_data = ""
    ssh_key_path = None
    if answers["answer"] == EXISTING_CONTENTS:
        print("Registering from file contents")
        ssh_key_data = input(
            "[\033[33m?\033[0m] Please paste the contents of your public ssh key. It should start with ssh-rsa: ")
    elif answers["answer"] == EXISTING_PATH:
        print("Register in vpc existing key from path")
        questions = [
            inquirer.Text(
                "public_key_path", message='Please paste path to your \033[92mpublic\033[0m ssh key', validate=validate_exists)
        ]
        answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
        with open(os.path.abspath(os.path.expanduser(answers["public_key_path"])), 'r') as file:
            ssh_key_data = file.read()
    else:
        ssh_key_data, ssh_key_path = generate_keypair(keyname)

    response = None
    try:    # regardless of the above, try registering an ssh-key 
        response = ibm_vpc_client.create_key(public_key=ssh_key_data, name=keyname, resource_group={
                                        "id": resource_group_id}, type='rsa')
    except ApiException as e:
        print(e)
            
        if "Key with name already exists" in e.message and keyname == DEFAULT_KEY_NAME:
            key = get_ssh_key(ibm_vpc_client, DEFAULT_KEY_NAME)
            ibm_vpc_client.delete_key(id=key['id'])
            response = ibm_vpc_client.create_key(public_key=ssh_key_data, name=keyname, resource_group={
                                        "id": resource_group_id}, type='rsa')                
        else: 
            if "Key with fingerprint already exists" in e.message:
                print(color_msg("Can't register an SSH key with the same fingerprint",Color.RED))
            raise # can't continue the configuration process without a valid ssh key       
            
    print(f"\033[92mnew SSH key {keyname} been registered in vpc\033[0m")

    result = response.get_result()
    return result['name'], result['id'], ssh_key_path


DEPENDENCIES = {'ibm_vpc': {'resource_group_id': None}}


class SshKeyConfig(ConfigBuilder):

    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        self.base_config = base_config

    def _validate_keypair(self, answers, current):
        if not current or not os.path.exists(os.path.abspath(os.path.expanduser(current))):
            raise errors.ValidationError(
            '', reason=f"File {current} doesn't exist")

        public_res = self.ibm_vpc_client.get_key(self.ssh_key_id).get_result()['public_key'].split(' ')[1]
        private_res = subprocess.getoutput([f"ssh-keygen -y -f {current} | cut -d' ' -f 2"])

        if not public_res == private_res:
            raise errors.ValidationError(
            '', reason=f"Private ssh key {current} and public key {self.ssh_key_name} are not a pair")

        return public_res == private_res

    @update_decorator
    def run(self) -> Dict[str, Any]:
        @spinner
        def get_ssh_key_objects():
            return self.ibm_vpc_client.list_keys().get_result()['keys']

        ssh_key_objects = get_ssh_key_objects()

        CREATE_NEW_SSH_KEY = "Register new SSH key in IBM VPC"

        default = find_default(self.defaults, ssh_key_objects, id='key_id')
        ssh_key_name, ssh_key_id = find_name_id(
            ssh_key_objects, 'Choose ssh key', do_nothing=CREATE_NEW_SSH_KEY, default=default)

        ssh_key_path = None
        if not ssh_key_name:
            ssh_key_name, ssh_key_id, ssh_key_path = register_ssh_key(
                self.ibm_vpc_client, self.base_config)

        self.ssh_key_id = ssh_key_id
        self.ssh_key_name = ssh_key_name

        if not ssh_key_path:
            questions = [
                inquirer.Text(
                    "private_key_path", message=f'Please paste path to \033[92mprivate\033[0m ssh key associated with selected public key {ssh_key_name}', validate=self._validate_keypair, default=self.defaults.get('ssh_key_filename') or "~/.ssh/id_rsa")
            ]
            answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
            ssh_key_path = os.path.abspath(
                os.path.expanduser(answers["private_key_path"]))

        # currently the user is hardcoded to root
        return ssh_key_id, ssh_key_path, 'root'

    @update_decorator
    def verify(self, base_config):
        default_keyname = DEFAULT_KEY_NAME
        
        if base_config.get('ibm_vpc'):
            resource_group_id = base_config['ibm_vpc']['resource_group_id']
        else:
            for available_node_type in base_config['available_node_types']:
                resource_group_id = base_config['available_node_types'][available_node_type]['node_config']['resource_group_id']
                break
        
        def is_pair(id, ssh_key_filename):
            public_res = self.ibm_vpc_client.get_key(id).get_result()['public_key'].split(' ')[1]
            private_res = subprocess.getoutput([f"ssh-keygen -y -f {ssh_key_filename} | cut -d' ' -f 2"])
            return public_res == private_res

        # user specified both vpc key id and private key, just validate they are a pair
        if self.defaults['key_id'] and self.defaults['ssh_key_filename']:
            if not is_pair(self.defaults['key_id'], self.defaults['ssh_key_filename']):
                raise errors.ValidationError(
                '', reason=f"Private ssh key {self.defaults['ssh_key_filename']} and public key {self.defaults['key_id']} are not a pair")
            
            return self.defaults['key_id'], self.defaults['ssh_key_filename'], 'root'
        elif self.defaults['ssh_key_filename']:
            # user provided private key only, check if there default ssh key already in vpc
            default_key = get_ssh_key(self.ibm_vpc_client, default_keyname)
            # and it match the specified private key
            if default_key and is_pair(default_key['id'], self.defaults['ssh_key_filename']):
                return default_key['id'], self.defaults['ssh_key_filename'], 'root'
            
            # not a pair. delete not matching default key from vpc if exists
            if default_key:
                self.ibm_vpc_client.delete_key(id=default_key['id'])
            # generate public key from private key
            ssh_key_data = subprocess.getoutput([f"ssh-keygen -y -f {self.defaults['ssh_key_filename']} | cut -d' ' -f 2"])
            response = self.ibm_vpc_client.create_key(public_key=ssh_key_data, name=default_keyname, resource_group={
                                        "id": resource_group_id}, type='rsa')
            print(f"\033[92mnew SSH key {default_keyname} been registered in vpc\033[0m")
            result = response.get_result()
            return result['id'], self.defaults['ssh_key_filename'], 'root'
        else:
            # check if there default ssh key already in vpc
            default_key = get_ssh_key(self.ibm_vpc_client, default_keyname)
            if default_key:
                self.ibm_vpc_client.delete_key(id=default_key['id'])

            # no ssh key information been provided by user, generate new keypair
            ssh_key_data, ssh_key_path = generate_keypair(default_keyname)    

            response = self.ibm_vpc_client.create_key(public_key=ssh_key_data, name=default_keyname, resource_group={
                                        "id": resource_group_id}, type='rsa')
            print(f"\033[92mnew SSH key {default_keyname} been registered in vpc\033[0m")
            result = response.get_result()
            return result['id'], ssh_key_path, 'root'

    @update_decorator
    def create_default(self):
        
        # if exist with same name - override
        _, key_id, ssh_key_path = register_ssh_key(self.ibm_vpc_client, self.base_config, auto=True)
        
        return key_id, ssh_key_path, 'root'


from ibm_ray_config.modules.config_builder import ConfigBuilder
from typing import Any, Dict

from ibm_ray_config.modules.utils import get_option_from_list

class FloatingIpConfig(ConfigBuilder):
    
    def run(self) -> Dict[str, Any]:
        head_ip = None
        floating_ips = self.ibm_vpc_client.list_floating_ips().get_result()['floating_ips']
        # filter away floating-ips that are already bound, or belong to a zone different than the one chosen by user  
        free_floating_ips = [ip for ip in floating_ips if not ip.get('target') and self.base_config['provider']['zone_name']==ip['zone']['name']] 

        if free_floating_ips:
            ALLOCATE_NEW_FLOATING_IP = 'Allocate new floating ip'
            head_ip_obj = get_option_from_list("Choose head ip", free_floating_ips, choice_key='address', do_nothing=ALLOCATE_NEW_FLOATING_IP)
            if head_ip_obj and (head_ip_obj != ALLOCATE_NEW_FLOATING_IP):
                head_ip = head_ip_obj['address']
                
            if self.base_config.get('available_node_types'):
                for available_node_type in self.base_config['available_node_types']:
                    if head_ip:
                        self.base_config['available_node_types'][available_node_type]['node_config']['head_ip'] = head_ip
                    else:
                        self.base_config['available_node_types'][available_node_type]['node_config'].pop('head_ip', None)
            else:
                self.base_config['available_node_types']['ray_head_default']['node_config']['head_ip'] = head_ip

        return self.base_config
    
    def create_default(self):
        NotImplementedError("This backend doesn't support it yet")



from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator, spinner
from typing import Any, Dict
from ibm_ray_config.modules.utils import find_obj, find_default


class ImageConfig(ConfigBuilder):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)

    @update_decorator
    def run(self) -> Dict[str, Any]:

        @spinner
        def get_image_objects():
            return self.ibm_vpc_client.list_images().get_result()['images']

        image_objects = get_image_objects()

        default = find_default({'name': 'ibm-ubuntu-20-04-'}, image_objects, name='name', substring=True)
        image_obj = find_obj(image_objects, 'Please choose \033[92mUbuntu\033[0m 20.04 VM image, currently only Ubuntu supported', default=default)

        return image_obj['id'], image_obj['minimum_provisioned_size'], image_obj['owner_type'] == 'user'

    @update_decorator
    def verify(self, base_config):
        image_id = self.defaults['image_id']
        image_objects = self.ibm_vpc_client.list_images().get_result()['images']
        if image_id:
            image_obj = find_obj(image_objects, 'dummy', obj_id=image_id)
        else:
            # find first occurrence
            image_obj = next((obj for obj in image_objects if 'ibm-ubuntu-20-04-' in obj['name']), None)
            
        return image_obj['id'], image_obj['minimum_provisioned_size'], image_obj['owner_type'] == 'user'

    @update_decorator
    def create_default(self):
        image_objects = self.ibm_vpc_client.list_images().get_result()['images']

        image_obj = next((image for image in image_objects if 'ibm-ubuntu-20-04-' in image['name']), None)
        
        print(f'Selected \033[92mUbuntu\033[0m 20.04 VM image, {image_obj["name"]}')
        return image_obj['id'], image_obj['minimum_provisioned_size'], image_obj['owner_type'] == 'user'



import importlib
import os
import re
import subprocess
import sys
import tempfile
import time
from enum import Enum
import inquirer
from inquirer import errors
from ibm_platform_services import IamIdentityV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
import ibm_cloud_sdk_core


CACHE = {}
ARG_STATUS = Enum('STATUS', 'VALID INVALID MISSING')  # variable possible status.

class MSG_STATUS(Enum):
    ERROR = '[ERROR]'
    WARNING = '[WARNING]'
    SUCCESS = '[SUCCESS]'

NEW_INSTANCE = 'Create a new'  # guaranteed substring in every 'create new instance' option prompted to user.


def get_option_from_list(msg, choices, default=None, choice_key='name', do_nothing=None, validate=True, carousel=True):
    """returns the user's choice out of given 'choices'. 
        :param msg - msg to display the user before presenting choices
        :param choices - list of choices to present user with
        :param default - default choice out of the lists of choices presented to user
        :param choice_key - if 'choices' is a dict `choice_key` represents the key associated with the desired values
        :param do_nothing:str - an additional choice for 'choices' to be appended to the list
        :param carousel - when set to true allows user to rotate from last to first choice and vice-a-versa    """
    if (len(choices) == 0 and do_nothing == None):
        error_msg = f"There no option for {msg}"
        print(error_msg)
        raise Exception(error_msg)

    if (len(choices) == 1 and not do_nothing):
        return (choices[0])

    if choice_key:
        choices_keys = [choice[choice_key] for choice in choices]
    else:
        choices_keys = [choice for choice in choices]

    if do_nothing:
        choices_keys.insert(0, do_nothing)

    questions = [
        inquirer.List('answer',
                      message=msg,
                      choices=choices_keys,
                      default=default,
                      validate=validate,
                      carousel=carousel,
                      ), ]
    answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)

    # now find the object by name in the list
    if answers['answer'] == do_nothing:
        return do_nothing
    else:
        if choice_key:
            return next((x for x in choices if x[choice_key] == answers['answer']), None)
        else:
            return next((x for x in choices if x == answers['answer']), None)


def inquire_user(msg, choices, default=None, choice_key='name', create_new_instance=None,
                 handle_strings=False, validate=True, carousel=True):
    
    """returns the user's choice out of given 'choices'. 
      :param str create_new_instance: when initialized adds a 'create' option that allows the user
                            to create an instance rather than to opt for one of the options.
      :param bool handle_strings: when set to True handles input of list of strings instead of list of dicts.
      :param str choice_key: creates options list presented to user (choices_keys) using this key """

    # options to be displayed to user
    choices_keys = [choice[choice_key] for choice in choices] if not handle_strings else choices

    if create_new_instance:
        choices_keys.insert(0, color_msg(create_new_instance, style=Style.ITALIC))

    if len(choices_keys) == 0:
        raise Exception(f"No options were found to satisfy the following request: {msg}")

    if len(choices_keys) == 1:
        if create_new_instance:
            print(color_msg(f"\nNo existing instances were found in relation to the request: "
                            f"'{msg}'. Create a new one to proceed. ", color=Color.RED))
        else:
            print(color_msg(f"single option was found in response to the request: '{msg}'."
                            f" \n{choices[0]} was automatically chosen\n", color=Color.LIGHTGREEN))
            return choices[0]

    questions = [
        inquirer.List('answer',
                      message=msg,
                      choices=choices_keys,
                      default=default,
                      validate=validate,
                      carousel=carousel,
                      )]
    answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)

    if create_new_instance and create_new_instance in answers['answer']:
        return create_new_instance
    elif handle_strings:  # returns the string the user chose
        return answers['answer']
    else:  # returns the object belonging to user's choice
        return next((x for x in choices if x[choice_key] == answers['answer']), None)

def find_obj(objects, msg, obj_id=None, obj_name=None, default=None, do_nothing=None):
    """returns object matching the specified keys' values: obj['id'] or obj['name']. Else, returns a user chosen object.   
    obj_id and obj_name are the values for: 'id' and 'name' - common keys in the API response format of IBM SDK.
    """
    obj = None
    if obj_id:
        # validating that obj exists
        obj = next((obj for obj in objects if obj['id'] == obj_id), None)
        if not obj:
            raise Exception(f'Object with specified id {obj_id} not found')
    if obj_name:
        obj = next((obj for obj in objects if obj['name'] == obj_name), None)
    
    if not obj:  # no matching values were found, collecting user's choice from 'objects'.
        obj = get_option_from_list(
            msg, objects, default=default, do_nothing=do_nothing)
    return obj

def find_name_id(objects, msg, obj_id=None, obj_name=None, default=None, do_nothing=None):
    obj = find_obj(objects, msg, obj_id, obj_name, default, do_nothing)
    if do_nothing and obj == do_nothing:
        return None, None

    return obj['name'], obj['id']

def validate_not_empty(answers, current):
    if not current:
        raise errors.ValidationError('', reason=f"Key name can't be empty")
    return True


def validate_exists(answers, current):
    if not current or not os.path.exists(os.path.abspath(os.path.expanduser(current))):
        raise errors.ValidationError(
            '', reason=f"File {current} doesn't exist")
    return True


def get_region_by_endpoint(endpoint):
    return re.search('//(.+?).iaas.cloud.ibm.com', endpoint).group(1)


def find_default(template_dict, objects, name=None, id=None, substring=False):
    """returns an object in 'objects' who's value for the field 'name' or 'id' equals that of item[name]/item[id] for an item in template_dict

    Args:
        template_dict (dict): dict from an existing configuration file that may contain desired values. 
        objects (list of dicts): list of dicts that may contain a key with the desired value.  
        name (str): name of the key of an object within 'template_dict' that contains the desired value, to be evaluated against obj['name'] of 'objects'.
        id (str): name of the key of an object within 'template_dict' that contains the desired value, to be evaluated against obj['id'] of 'objects'
        substring (bool): if set to true, the substring value of the desired val in template_dict is tested against the value of 'objects' 
    """

    val = None
    for k, v in template_dict.items():
        if isinstance(v, dict):
            return find_default(v, objects, name=name, id=id)
        else:
            if name:
                key = 'name'
                if k == name:
                    val = v
            elif id:
                key = 'id'
                if k == id:
                    val = v

            if val:
                if not substring:
                    obj = next((obj for obj in objects if obj[key] == val), None)
                else:
                    obj = next((obj for obj in objects if val in obj[key]), None)
                if obj:
                    return obj['name']


def free_dialog(msg, default=None, validate=True):
    """ Returns user's answer to an open-ended question """
    question = [
        inquirer.Text('answer',
                      message=msg,
                      default=default,
                      validate=validate)]
    answer = inquirer.prompt(question, raise_keyboard_interrupt=True)
    return answer


def password_dialog(msg, default=None, validate=True):
    question = [
        inquirer.Password('answer',
                          message=msg,
                          default=default,
                          validate=validate)]
    answer = inquirer.prompt(question, raise_keyboard_interrupt=True)
    return answer


def get_confirmation(msg, default=None):
    questions = [
        inquirer.Confirm('answer',
                         message=msg,
                         default=default,
                         ), ]
    answer = inquirer.prompt(questions, raise_keyboard_interrupt=True)

    return answer


def retry_on_except(retries, sleep_duration, error_msg='', default=None):
    """A decorator that calls the decorated function up to a number of times equals to 'retires' with a given
      'sleep_duration' in between.
       if the function failed the allotted number of retries, a default value will be returned,
       granted that one was provided, else the config tool will be terminated. """

    def retry_on_except_warpper(func):
        def func_wrapper(*args, **kwargs):
            msg = error_msg  # transferring the value via a mediator is necessary due to decorator's restrictions.
            for retry in range(retries):
                try:
                    result = func(*args, **kwargs)
                    return result

                except Exception as e:
                    msg += str(e)
                    if retry < retries - 1:  # avoid sleeping after last failure
                        time.sleep(sleep_duration)
            if default:
                print(color_msg(f"{msg}", color=Color.RED))
                return default
            else:
                print(color_msg(f"{msg}\nConfig tool was terminated", color=Color.RED))
                sys.exit(1)

        return func_wrapper

    return retry_on_except_warpper


def run_cmd(cmd):
    """runs a command via cli while constantly printing the output from the read pipe"""

    # small buffer size forces the process not to buffer the output, thus printing it instantly.
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1, shell=True)
    for line in iter(process.stdout.readline, b''):
        print(line.decode())
    process.stdout.close()
    process.wait()


def verify_paths(input_path, output_path, verify_config=False):
    """:returns a valid input and output path files, in accordance with provided paths.
        if a given path is invalid, and user is unable to rectify, a default path will be chosen in its stead. """

    def _is_valid_input_path(path):
        if not os.path.isfile(path):
            print(color_msg(f"\nError - Path: '{path}' doesn't point to a file. ", color=Color.RED))
            return False
        return True

    def _is_valid_output_path(path):
        """:returns path if it's either a valid absolute path, or a file name to be appended to current directory"""
        dir_file = path.rsplit('/', 1)
        prefix_directory = dir_file[0]
        if len(dir_file) == 1 or os.path.isdir(prefix_directory):
            return path
        else:
            print(color_msg(f"{prefix_directory} doesn't lead to an existing directory", color=Color.RED))

    def _prompt_user(path, default_config_file, verify_func, request, default_msg):
        while True:
            if not path:
                print(color_msg(default_msg, color=Color.LIGHTGREEN))
                return default_config_file

            if verify_func(path):
                return path
            else:
                path = free_dialog(request)['answer']

    if not verify_config:
        input_path = _prompt_user(input_path, '', _is_valid_input_path,
                                  "Provide a path to your existing config file, or leave blank to configure from template",
                                  'Using default input file\n')
    output_path = _prompt_user(output_path, tempfile.mkstemp(suffix='.yaml')[1], _is_valid_output_path,
                               "Provide a custom path for your config file, or leave blank for default output location",
                               'Using default output path\n')
    return input_path, output_path

def verify_iam_api_key(answers, apikey, iam_endpoint=None):
    """Terminates the config tool if no IAM_API_KEY matching the provided value exists"""

    iam_identity_service = IamIdentityV1(authenticator=IAMAuthenticator(apikey, url=iam_endpoint))
    try:
        iam_identity_service.get_api_keys_details(iam_api_key=apikey)
    except ibm_cloud_sdk_core.api_exception.ApiException:
        raise errors.ValidationError('', reason=color_msg(f"No IAmApiKey matching the given value {apikey} was found.", Color.RED))
    return True

def color_msg(msg, color=None, style=None, background=None):
    """reformat a given string and returns it, matching the desired color,style and background in Ansi color code.
        parameters are Enums of the classes: Color, Style and Background."""

    init = '\033['
    end = '\033[m'
    font = ''
    if color:
        font += color.value
    if style:
        font = font + ';' if font else font
        font += style.value

    if background:
        font = font + ';' if font else font
        font += background.value

    return init + font + 'm' + msg + end


class Color(Enum):
    BLACK = '30'
    RED = '31'
    GREEN = '32'
    BROWN = '33'
    BLUE = '34'
    PURPLE = '35'
    CYAN = '36'
    LIGHTGREY = '37'
    DARKGREY = '90'
    LIGHTRED = '91'
    LIGHTGREEN = '92'
    YELLOW = '93'
    LIGHTBLUE = '94'
    PINK = '95'
    LIGHTCYAN = '96'


class Style(Enum):
    RESET = '0'
    BOLD = '01'
    DISABLE = '02'
    ITALIC = '03'
    UNDERLINE = '04'
    REVERSE = '07'
    STRIKETHROUGH = '09'


class Background(Enum):
    BLACK = '40'
    RED = '41'
    GREEN = '42'
    ORANGE = '43'
    BLUE = '44'
    PURPLE = '45'
    CYAN = '46'
    LIGHTGREY = '47'



from typing import Any, Dict
from ibm_ray_config.modules.api_key import ApiKeyConfig
from ibm_ray_config.modules.config_builder import ConfigBuilder
from ibm_ray_config.modules.utils import verify_iam_api_key

class RayApiKeyConfig(ApiKeyConfig):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        self.base_config.pop('ibm', None)
        self.defaults['api_key'] = self.base_config['provider']['iam_api_key'] if self.base_config.setdefault('provider', {}) \
            else None

    def update_config(self, iam_api_key, compute_iam_endpoint=None):
        self.base_config['provider']['iam_api_key'] = iam_api_key

        if compute_iam_endpoint:
            self.base_config['provider']['iam_endpoint'] = compute_iam_endpoint

        self.base_config.pop('ibm', None)
        return self.base_config

    def verify(self, base_config):
        api_key = base_config['provider']['iam_api_key']

        verify_iam_api_key(None, api_key)
        ConfigBuilder.iam_api_key = api_key

        return base_config



from typing import Any, Dict
from ibm_ray_config.modules.gen2.image import ImageConfig

class RayImageConfig(ImageConfig):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        
        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.defaults['image_id'] = self.base_config['available_node_types'][available_node_type]['node_config'].get('image_id')
                break
    
    def update_config(self, image_id, minimum_provisioned_size, custom_image):
        #minimum_provisioned_size will be used once non default image used
        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.base_config['available_node_types'][available_node_type]['node_config']['image_id'] = image_id
                self.base_config['available_node_types'][available_node_type]['node_config']['boot_volume_capacity'] = minimum_provisioned_size
        else:
            self.base_config['available_node_types']['ray_head_default']['node_config']['image_id'] = image_id
            self.base_config['available_node_types']['ray_head_default']['node_config']['boot_volume_capacity'] = minimum_provisioned_size

        # if custom image, all setup commands should be removed
        if custom_image:
            self.base_config['setup_commands'] = ['rm -f ~/.ray/tags.json']



import time

from ibm_cloud_sdk_core import ApiException
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_vpc import VpcV1


def delete_config(config):

    # parse config
    if 'provider' in config:
        from ibm_ray_config.modules.gen2.ray import parse_config
    else:
        raise Exception('Config file not supported')
    
    vpc_config = parse_config(config)
    
    authenticator = IAMAuthenticator(vpc_config['iam_api_key'],
                                     url=vpc_config.get('iam_endpoint'))
    ibm_vpc_client = VpcV1('2021-01-19', authenticator=authenticator)
    ibm_vpc_client.set_service_url(vpc_config['endpoint'] + '/v1')
    
    # find and delete all vpc vsis
    instances_info = ibm_vpc_client.list_instances(vpc_id=vpc_config['vpc_id']).get_result()
    for ins in instances_info['instances']:
        # delete floating ips
        print('Deleting instance {}'.format(ins['name']))
        
        interface_id = ins['network_interfaces'][0]['id']
        fips = ibm_vpc_client.list_instance_network_interface_floating_ips(
                    ins['id'], interface_id).get_result()['floating_ips']
        if fips:
            fip = fips[0]['id']
            ibm_vpc_client.delete_floating_ip(fip)
        
        # delete instance        
        ibm_vpc_client.delete_instance(ins['id'])

    time.sleep(5)
    
    # delete subnet
    print('Deleting subnet')
    try:
        ibm_vpc_client.delete_subnet(vpc_config['subnet_id'])
    except ApiException as e:
        if e.code == 404:
            pass
        else:
            raise e

        time.sleep(25)

    # delete gateway?
    print('Deleting gateway')
    gateways = ibm_vpc_client.list_public_gateways().get_result()
    for gw in gateways['public_gateways']:
        if gw['vpc']['id'] == vpc_config['vpc_id']:
            ibm_vpc_client.delete_public_gateway(gw['id'])

    time.sleep(15)
    
    # delete ssh key
    try:
        ibm_vpc_client.delete_key(id=vpc_config['key_id'])
    except ApiException as e:
        if e.code == 404:
            pass
        else:
            raise e

        time.sleep(5)
    
    # delete vpc
    print('Deleting VPC')
    try:
        ibm_vpc_client.delete_vpc(vpc_config['vpc_id'])
    except ApiException as e:
        if e.code == 404:
            pass
        else:
            raise e



from typing import Any, Dict

import inquirer
from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator, spinner
from ibm_ray_config.modules.utils import (find_default, find_name_id,
                                        get_option_from_list,
                                        get_region_by_endpoint,
                                        validate_not_empty, CACHE)


class VPCConfig(ConfigBuilder):

    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        self.base_config = base_config

        self.sg_rules = {}  # security group rules.
        self.vpc_name = 'cluster-vpc'
        
    def _get_region(self):
        region = None
        try:
            region = get_region_by_endpoint(self.ibm_vpc_client.service_url)
        except Exception:
            region = ConfigBuilder.region
        return region

    @update_decorator
    def run(self) -> Dict[str, Any]:
        region = self._get_region()

        vpc_obj, zone_obj = self._select_vpc(self.base_config, region)

        if not vpc_obj:
            raise Exception(f'Failed to select VPC')

        all_subnet_objects = self.ibm_vpc_client.list_subnets().get_result()[
            'subnets']

        # filter only subnets from selected availability zone
        subnet_objects = [s_obj for s_obj in all_subnet_objects if s_obj['zone']
                          ['name'] == zone_obj['name'] and s_obj['vpc']['id'] == vpc_obj['id']]

        if not subnet_objects:
            raise f'Failed to find subnet for vpc {vpc_obj["name"]} in zone {zone_obj["name"]}'

        return vpc_obj, zone_obj, subnet_objects[0]['id']

    def _build_security_group_rule_prototype_model(self, missing_rule, sg_id=None):
        direction, protocol, port = missing_rule.split('_')
        remote = {"cidr_block": "0.0.0.0/0"}

        try: # port number was specified
            port = int(port)
            port_min = port
            port_max = port
        except:
            port_min = 1
            port_max = 65535

            # only valid if security group already exists
            if port == 'sg':
                if not sg_id:
                    return None
                remote = {'id': sg_id}

        return {
            'direction': direction,
            'ip_version': 'ipv4',
            'protocol': protocol,
            'remote': remote,
            'port_min': port_min,
            'port_max': port_max
        }

    def _create_vpc(self, ibm_vpc_client, resource_group, vpc_default_name, auto=False):

        q = [
            inquirer.Text('name', message="Please, type a name for the new VPC",
                          validate=validate_not_empty, default=vpc_default_name),
            inquirer.List('answer', message='Create new VPC and configure required rules in default security group', choices=[
                          'yes', 'no'], default='yes')
        ]

        answers = {'answer': 'yes',
                   'name': vpc_default_name}
        if not auto:    # use default answers, i.e. create a new VPC and choose default VPC name
            answers = inquirer.prompt(q, raise_keyboard_interrupt=True)
        
        if answers['answer'] == 'yes':  # create a new VPC

            @spinner
            def _create():
                return ibm_vpc_client.create_vpc(address_prefix_management='auto', classic_access=False,
                                                name=answers['name'], resource_group=resource_group).get_result()

            return _create()
        else:
            return None

    def _create_vpc_peripherals(self, ibm_vpc_client, vpc_obj, zone_obj, resource_group):
        vpc_name = vpc_obj['name']
        vpc_id = vpc_obj['id']
        # create subnet
        subnet_name = '{}-subnet'.format(vpc_name)
        subnet_data = None

        # find cidr
        ipv4_cidr_block = None
        res = ibm_vpc_client.list_vpc_address_prefixes(
            vpc_id).result
        address_prefixes = res['address_prefixes']

        # searching for the CIDR block (internal ip range) matching the specified zone of a VPC (whose region has already been set) 
        for address_prefix in address_prefixes:
            if address_prefix['zone']['name'] == zone_obj['name']:
                ipv4_cidr_block = address_prefix['cidr']
                break

        subnet_prototype = {}
        subnet_prototype['zone'] = {'name': zone_obj['name']}
        subnet_prototype['ip_version'] = 'ipv4'
        subnet_prototype['name'] = subnet_name
        subnet_prototype['resource_group'] = resource_group
        subnet_prototype['vpc'] = {'id': vpc_id}
        subnet_prototype['ipv4_cidr_block'] = ipv4_cidr_block

        subnet_data = ibm_vpc_client.create_subnet(
            subnet_prototype).result
        subnet_id = subnet_data['id']

        # create public gateway
        gateway_id = self.create_public_gateway(vpc_obj, zone_obj, resource_group, subnet_name)

        # Attach public gateway to the subnet
        ibm_vpc_client.set_subnet_public_gateway(
            subnet_id, {'id': gateway_id})

        print(
            f"\033[92mVPC subnet {subnet_prototype['name']} been created and attached to gateway\033[0m")

        # Update security group to have all required rules
        sg_id = vpc_obj['default_security_group']['id']

        # update sg name
        sg_name = '{}-sg'.format(vpc_name)
        ibm_vpc_client.update_security_group(
            sg_id, security_group_patch={'name': sg_name})

        # add rule to open private tcp traffic between VSIs within the security group
        sg_rule_prototype = self._build_security_group_rule_prototype_model(
            'inbound_tcp_sg', sg_id=sg_id)
        res = ibm_vpc_client.create_security_group_rule(
            sg_id, sg_rule_prototype).get_result()

        # add all other required rules configured by the specific backend
        for rule in self.sg_rules.keys():
            sg_rule_prototype = self._build_security_group_rule_prototype_model(
                rule)
            if sg_rule_prototype:
                res = ibm_vpc_client.create_security_group_rule(
                    sg_id, sg_rule_prototype).get_result()

        print(
            f"\033[92mSecurity group {sg_name} been updated with required rules\033[0m\n")

    def create_public_gateway(self, vpc_obj, zone_obj, resource_group, subnet_name):
        vpc_id = vpc_obj['id']
        
        gateway_prototype = {}
        gateway_prototype['vpc'] = {'id': vpc_id}
        gateway_prototype['zone'] = {'name': zone_obj['name']}
        gateway_prototype['name'] = f"{subnet_name}-gw"
        gateway_prototype['resource_group'] = resource_group
        gateway_data = self.ibm_vpc_client.create_public_gateway(
            **gateway_prototype).get_result()
        gateway_id = gateway_data['id']

        print(
            f"\033[92mVPC public gateway {gateway_prototype['name']} been created\033[0m")

        return gateway_id

    def _select_resource_group(self, auto=False):
        # find resource group

        @spinner
        def get_res_group_objects():
            return self.resource_service_client.list_resource_groups().get_result()['resources']

        res_group_objects = get_res_group_objects()
        if auto:
            return res_group_objects[0]['id']

        default = find_default(
            self.defaults, res_group_objects, id='resource_group_id')
        res_group_obj = get_option_from_list(
            "Select resource group", res_group_objects, default=default)
        return res_group_obj['id']

    def _select_zone(self, vpc_id, region, auto=False):
        # find availability zone
        @spinner
        def get_zones_and_subnets():
            zones_objects = self.ibm_vpc_client.list_region_zones(region).get_result()['zones']
            all_subnet_objects = self.ibm_vpc_client.list_subnets().get_result()['subnets']
            return zones_objects, all_subnet_objects

        zones_objects, all_subnet_objects = get_zones_and_subnets()
        
        if auto:
            return zones_objects[0]

        if vpc_id:
            # filter out zones that given vpc has no subnets in
            zones = [s_obj['zone']['name']
                        for s_obj in all_subnet_objects if s_obj['vpc']['id'] == vpc_id]
            zones_objects = [
                z for z in zones_objects if z['name'] in zones]

        try:
            default = find_default(
                self.defaults, zones_objects, name='zone_name')
            zone_obj = get_option_from_list(
                "Choose availability zone", zones_objects, default=default)
        except:
            raise Exception(
                "Failed to list zones for selected vpc {vpc_id}, please check whether vpc missing subnet")

        return zone_obj

    def _select_vpc(self, node_config, region):

        vpc_id, vpc_name, zone_obj, sg_id, resource_group = None, None, None, None, None
        ibm_vpc_client = self.ibm_vpc_client
       
        while True:
            CREATE_NEW = 'Create new VPC'

            @spinner
            def list_vpcs():
                return ibm_vpc_client.list_vpcs().get_result()['vpcs']

            vpc_objects = list_vpcs()
            default = find_default(self.defaults, vpc_objects, id='vpc_id')

            vpc_name, vpc_id = find_name_id(
                vpc_objects, "Select VPC", obj_id=vpc_id, do_nothing=CREATE_NEW, default=default)

            zone_obj = self._select_zone(vpc_id, region)

            # User didn't choose an existing VPC 
            if not vpc_name:    
                resource_group_id = self._select_resource_group()
                resource_group = {'id': resource_group_id}

                # Create a unique VPC name 
                vpc_default_name = self.vpc_name
                c = 1
                vpc_names = [vpc_obj['name'] for vpc_obj in vpc_objects]
                while vpc_default_name in vpc_names:    # find next default available vpc name
                    vpc_default_name = f'{self.vpc_name}-{c}'
                    c += 1

                vpc_obj = self._create_vpc(ibm_vpc_client, resource_group, vpc_default_name)
                if not vpc_obj:     # User declined to create a new VPC
                    continue
                else:   # User created a new VPC
                    vpc_name = vpc_obj['name']
                    vpc_id = vpc_obj['id']

                    print(f"\n\n\033[92mVPC {vpc_name} been created\033[0m")

                    self._create_vpc_peripherals(ibm_vpc_client, vpc_obj, zone_obj, resource_group)
                    break

            # User chose an existing VPC 
            else:
                # validate chosen vpc has all requirements
                # starting from validating each of its subnets has public gateway

                @spinner
                def get_vpc_obj_and_subnets():
                    vpc_obj = ibm_vpc_client.get_vpc(id=vpc_id).result
                    all_subnet_objects = ibm_vpc_client.list_subnets().get_result()['subnets']
                    return vpc_obj, all_subnet_objects

                vpc_obj, all_subnet_objects = get_vpc_obj_and_subnets()
                resource_group = {'id': vpc_obj['resource_group']['id']}

                subnet_objects = [s_obj for s_obj in all_subnet_objects if s_obj['zone']
                          ['name'] == zone_obj['name'] and s_obj['vpc']['id'] == vpc_obj['id']]

                gw_id = None
                for subnet in subnet_objects:
                    gw = subnet.get('public_gateway')        
                    if not gw:
                        if not gw_id:
                            questions = [
                                inquirer.List('answer',
                                    message=f'Selected vpcs {vpc_obj["name"]} subnet {subnet["name"]} is missing required public gateway, create a new one?',
                                    choices=['yes', 'no'], default='yes'
                                ), ]

                            answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)

                            if answers['answer'] == 'yes':
                                gw_id = self.create_public_gateway(vpc_obj, zone_obj, resource_group, subnet['name'])
                            else:
                                exit(1)

                        # attach gateway to subnet
                        ibm_vpc_client.set_subnet_public_gateway(subnet['id'], {'id': gw_id})
                    else:
                        gw_id = gw['id']
                break

        CACHE['resource_group_id'] = resource_group['id']
        
        return vpc_obj, zone_obj

    @update_decorator
    def verify(self, base_config):
        # if vpc_id not specified will look for the first one

        if self.defaults['vpc_id']:
            vpc_obj = self.ibm_vpc_client.get_vpc(id=self.defaults['vpc_id']).result
        else:
            vpc_objects = self.ibm_vpc_client.list_vpcs().result
            if vpc_objects['total_count'] > 0:
                # return first vpc occurrence
                vpc_obj = vpc_objects['vpcs'][0]
            else:
                # create new vpc
                res_group_objects = self.resource_service_client.list_resource_groups().get_result()['resources']
            
                print(f"Selected first found resource group {res_group_objects[0]['name']}")
                resource_group = resource_group = {'id': res_group_objects[0]['id']}
            
                region = self._get_region()
                print(f"\n\n\033[92mRegion {region} been selected\033[0m")
                        
                vpc_obj = self.ibm_vpc_client.create_vpc(address_prefix_management='auto', classic_access=False,
                                        name="ray-default-vpc", resource_group=resource_group).get_result()
            
                print(f"\n\n\033[92mVPC {vpc_obj['name']} been created\033[0m")
            
                zones_objects = self.ibm_vpc_client.list_region_zones(region).get_result()['zones']
                zone_obj = zones_objects[0]
                self._create_vpc_peripherals(self.ibm_vpc_client, vpc_obj, zone_obj, resource_group)

        subnet_objects = self.ibm_vpc_client.list_subnets().get_result()['subnets']
        return vpc_obj, subnet_objects[0]['zone'], subnet_objects[0]['id']
    
    @update_decorator
    def create_default(self):
        region = self._get_region()
        resource_group_id = self._select_resource_group(auto=True)
        resource_group = {'id': resource_group_id}

        vpc_objects = self.ibm_vpc_client.list_vpcs().get_result()['vpcs']
        vpc_obj = next((vpc_obj for vpc_obj in vpc_objects if vpc_obj['name'] == self.vpc_name), None)
        
        if vpc_obj:
            # TODO: validate existing
            print(f"\n\n\033[92mUsing existing VPC with default name {vpc_obj['name']} \033[0m")
        else:
            vpc_obj = self._create_vpc(self.ibm_vpc_client,
                                    resource_group, self.vpc_name, auto=True)
            if not vpc_obj:
                raise Exception(f"Failed to create VPC {self.vpc_name}")
            else:
                print(f"\n\n\033[92mVPC {vpc_obj['name']} been created\033[0m")

            zone_obj = self._select_zone(vpc_obj['id'], region, auto=True)
            self._create_vpc_peripherals(self.ibm_vpc_client,
                                        vpc_obj,
                                        zone_obj,
                                        resource_group)
        
        zone_obj = self._select_zone(vpc_obj['id'], region, auto=True)
        CACHE['resource_group_id'] = resource_group['id']
        
        all_subnet_objects = self.ibm_vpc_client.list_subnets().get_result()[
            'subnets']

        # filter only subnets from selected availability zone
        subnet_objects = [s_obj for s_obj in all_subnet_objects if s_obj['zone']
                          ['name'] == zone_obj['name'] and s_obj['vpc']['id'] == vpc_obj['id']]

        if not subnet_objects:
            raise f'Failed to find subnet for vpc {vpc_obj["name"]} in zone {zone_obj["name"]}'

        return vpc_obj, zone_obj, subnet_objects[0]['id']



import importlib
import pkg_resources
import os
import click
import yaml

from ibm_ray_config.modules.utils import color_msg, Color, verify_paths

# currently supporting configuration for ray above IBM Gen2 VPC. 
IBM_VPC = 'IBM VPC'
backends =[{'name': IBM_VPC, 'path': 'gen2.ray'}]

def select_backend(input_file, default_config_suffix=''):
    """returns a base config file for the backend and its package """
    backend = backends[0] # currently supporting a single backend
   
    base_config = {}

    if input_file:
        with open(input_file) as f:
            base_config = yaml.safe_load(f)
    else:
        base_config = load_base_config(backend, default_config_suffix)

    # find the right modules
    backend_pkg = importlib.import_module(f"ibm_ray_config.modules.{backend['path']}")

    return base_config, backend_pkg

def load_base_config(backend, default_config_suffix=''):
    """returns a default base configuration dict for the chosen backend """
    backend_path = backend['path'].replace('.', '/')
    dir_path = os.path.dirname(os.path.realpath(__file__))
    input_file = f"{dir_path}/modules/{backend_path}/defaults{default_config_suffix}.yaml"

    base_config = None
    with open(input_file) as f:
        base_config = yaml.safe_load(f)

    return base_config

def validate_api_keys(base_config, modules, iam_api_key, compute_iam_endpoint):
    """validates the api key specified.
    returns a base config dict, updated with the api-key field populated and the post api-key module removal modules list  """
    # The API_KEY module is invoked and popped from the list. 
    api_key_module = modules[0]
    base_config = api_key_module(base_config).run(api_key=iam_api_key,
                                                  compute_iam_endpoint=compute_iam_endpoint)

    modules = modules[1:]
    return base_config, modules
    
@click.command()
@click.option('--output-file', '-o', help='Output filename to save configurations')
@click.option('--input-file', '-i', help=f'Template for the new configuration')
@click.option('--iam-api-key', '-a', help='IAM_API_KEY')
@click.option('--version', '-v', help=f'Get package version', is_flag=True)
@click.option('--compute-iam-endpoint', help='IAM endpoint url used for compute instead of default https://iam.cloud.ibm.com')
@click.option('--endpoint', help='IBM Cloud API endpoint')
@click.option('--pr', '-g', help=f'Temporary workaround for ray gen2 only. If specified, use provider setup from PR github', is_flag=True, default=False)
def builder(iam_api_key, output_file, input_file, version, compute_iam_endpoint, endpoint, pr):
    defaults = None  # to be replaced by a flag  
    if version:
        print(f"{pkg_resources.get_distribution('ibm-ray-config').project_name} "
              f"{pkg_resources.get_distribution('ibm-ray-config').version}")
        exit(0)

    print(color_msg("\nWelcome to ibm_ray_config export helper\n", color=Color.YELLOW))

    input_file, output_file = verify_paths(input_file, output_file)

    default_config_suffix = ''
    if pr:
        default_config_suffix = '_pr'

    base_config, backend_pkg = select_backend(input_file, default_config_suffix)
    
    modules = backend_pkg.MODULES
    base_config['create_defaults'] = defaults
    base_config, modules = validate_api_keys(base_config, modules, iam_api_key, compute_iam_endpoint)

    if endpoint:
        base_config['provider']['endpoint'] = endpoint

    for module in modules:
        next_module = module(base_config)
        
        if defaults:
            base_config = next_module.create_default()
        else:
            base_config = next_module.run()

    with open(output_file, 'w') as outfile:
        del base_config['create_defaults']
        yaml.dump(base_config, outfile, default_flow_style=False)

    if hasattr(backend_pkg, 'finish_message'):
        print(backend_pkg.finish_message(output_file))
    else:
        print("\n\n=================================================")
        print(color_msg(f"Cluster config file: {output_file}", color=Color.LIGHTGREEN))
        print("=================================================")

def error(msg):
    print(msg)
    raise Exception(msg)

def generate_config(*args, **kwargs):
    """A programmatic avenue to create configuration files. To be used externally by user."""
    backend = backends[0] # currently supporting a single backend
    _, output_file = verify_paths(None, None)
    
    # now update base config with backend specific params
    base_config = importlib.import_module(f"ibm_ray_config.modules.{backend['path']}").load_config(backend, *args, **kwargs)
    
    # now find the right modules
    modules = importlib.import_module(f"ibm_ray_config.modules.{backend['path']}").MODULES
    
    for module in modules:
        base_config = module(base_config).verify(base_config)

    with open(output_file, 'w') as outfile:
        yaml.dump(base_config, outfile, default_flow_style=False)

    print("\n\n=================================================")
    print(color_msg(f"Cluster config file: {output_file}", color=Color.LIGHTGREEN))
    print("=================================================")

    return output_file

def delete_config(config_file_path):
    config = None
    with open(config_file_path) as f:
        config = yaml.safe_load(f)
        
    from ibm_ray_config.modules.gen2 import delete_config
    delete_config(config)      
    
if __name__ == '__main__':
    try:
        builder()
    except KeyboardInterrupt:
        # User interrupt the program
        exit()



import logging
from typing import Any, Dict
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_platform_services import ResourceControllerV2, ResourceManagerV2
from ibm_vpc import VpcV1
from ibm_watson import IAMTokenManager
import threading
import time
import sys
from ibm_ray_config.modules.utils import CACHE, find_default, get_option_from_list

logger = logging.getLogger(__name__)


def update_decorator(f):
    def foo(*args, **kwargs):
        result = f(*args, **kwargs)
        update_config = getattr(args[0], 'update_config')
        if not result:
            return args[0].base_config
        if isinstance(result, tuple):
            update_config(*result)
        else:
            update_config(result)
        return args[0].base_config

    return foo


class ConfigBuilder:
    """
    Interface for building IBM Cloud config files for Ray
    """
    iam_api_key, ibm_vpc_client, resource_service_client, resource_controller_service, compute_iam_endpoint, region = None, None, None, None, None, None

    def __init__(self, base_config: Dict[str, Any]) -> None:

        self.defaults = {}
        if not ConfigBuilder.iam_api_key:
            if 'ibm' in base_config and 'iam_api_key' in base_config['ibm']:
                ConfigBuilder.iam_api_key = base_config['ibm']['iam_api_key']
            elif 'provider' in base_config and 'iam_api_key' in base_config['provider']:
                ConfigBuilder.iam_api_key = base_config['provider']['iam_api_key']

        if not ConfigBuilder.ibm_vpc_client and ConfigBuilder.iam_api_key:
            authenticator = IAMAuthenticator(ConfigBuilder.iam_api_key, url=ConfigBuilder.compute_iam_endpoint)
            ConfigBuilder.ibm_vpc_client = VpcV1(
                '2021-01-19', authenticator=authenticator)
            ConfigBuilder.resource_service_client = ResourceManagerV2(
                authenticator=authenticator)
            ConfigBuilder.resource_controller_service = ResourceControllerV2(
                authenticator=authenticator)

        self.init_clients(ConfigBuilder.iam_api_key, ConfigBuilder.compute_iam_endpoint)

        self.base_config = base_config

    def init_clients(self, iam_api_key, iam_endpoint=None):
        authenticator = IAMAuthenticator(iam_api_key, url=iam_endpoint)
        self.ibm_vpc_client = ConfigBuilder.ibm_vpc_client
        self.resource_service_client = ResourceManagerV2(authenticator=authenticator)
        self.resource_controller_service = ResourceControllerV2(authenticator=authenticator)


    """Interacts with user to get all required parameters"""

    def run(self, config) -> Dict[str, Any]:
        """Return updated config dictionary that can be dumped to config file

        Run interactive questionnaire
        """
        raise NotImplementedError

    """Updates specified config dictionary"""

    def update_config(self, *args) -> Dict[str, Any]:
        """Updates config dictionary that can be dumped to config file"""
        return self.base_config

    def get_resources(self, resource_type=None):
        """
        :param resource_type: str of the following possible values: ['service_instance','resource_instance']
        :return: resources belonging to a specific resource group, filtered by provided resource_type
        """

        if 'resource_group_id' not in CACHE:
            self.select_resource_group()

        @spinner
        def _get_resources():
            res = self.resource_controller_service.list_resource_instances(
                resource_group_id=CACHE['resource_group_id'], type=resource_type).get_result()
            resource_instances = res['resources']

            while res['next_url']:
                start = res['next_url'].split('start=')[1]
                res = self.resource_controller_service.list_resource_instances(
                    resource_group_id=CACHE['resource_group_id'], type=resource_type,
                    start=start).get_result()

                resource_instances.extend(res['resources'])
            return resource_instances

        return _get_resources()

    def select_resource_group(self):
        """returns resource group id of a resource group the user will be prompted to pick.
        stores result in CACHE['resource_group_id'] for further usage"""

        @spinner
        def get_resource_groups():
            return self.resource_service_client.list_resource_groups().get_result()['resources']

        res_group_objects = get_resource_groups()

        default = find_default(self.defaults, res_group_objects, id='resource_group_id')
        res_group_obj = get_option_from_list("Select resource group", res_group_objects, default=default)

        CACHE['resource_group_id'] = res_group_obj['id']  # cache group resource id for later use in storage

        return res_group_obj['id']

    def get_oauth_token(self):
        """:returns a temporary authentication token required by various IBM cloud APIs """

        iam_token_manager = IAMTokenManager(apikey=self.base_config['ibm']['iam_api_key'], url=ConfigBuilder.compute_iam_endpoint)
        return iam_token_manager.get_token()

    @update_decorator
    def verify(self, base_config):
        pass


class Spinner(threading.Thread):

    def __init__(self, *args, **kwargs):
        super(Spinner, self).__init__(*args, **kwargs)
        self.sttop = False

    def stop(self):
        self.sttop = True

    def stopped(self):
        return self.sttop

    def run(self):
        while True:
            if self.stopped():
                sys.stdout.write('\b')
                sys.stdout.flush()
                return

            for cursor in '\\|/-':
                time.sleep(0.1)
                sys.stdout.write('\r{}'.format(cursor))
                sys.stdout.flush()


def spinner(f):
    def foo(*args, **kwargs):
        s = Spinner()
        s.daemon = True
        s.start()
        result = f(*args, **kwargs)

        s.stop()
        s.join()

        return result

    return foo



import re
from typing import Any, Dict

import inquirer
from ibm_ray_config.modules.config_builder import ConfigBuilder
from ibm_ray_config.modules.utils import find_default, get_option_from_list


class WorkersConfig(ConfigBuilder):

    def run(self) -> Dict[str, Any]:
        default_cluster_name = self.base_config.get('cluster_name', 'default')
        default_min_workers = self.base_config.get('min_workers', '0')
        default_max_workers = default_min_workers

        question = [
            inquirer.Text(
                'name', message="Cluster name, either leave default or type a new one", default=default_cluster_name),
            inquirer.Text('min_workers', message="Minimum number of worker nodes",
                          default=default_min_workers, validate=lambda _, x: re.match('^[+]?[0-9]+$', x)),
            inquirer.Text('max_workers', message="Maximum number of worker nodes", default=default_max_workers,
                          validate=lambda answers, x: re.match('^[+]?[0-9]+$', x) and int(x) >= int(answers['min_workers']))
        ]

        answers = inquirer.prompt(question, raise_keyboard_interrupt=True)
        self.base_config['cluster_name'] = answers['name']
        self.base_config['max_workers'] = int(answers['max_workers'])

        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.base_config['available_node_types'][available_node_type]['min_workers'] = int(
                    answers['min_workers'])
                self.base_config['available_node_types'][available_node_type]['max_workers'] = int(
                    answers['max_workers'])
        else:
            self.base_config['available_node_types']['ray_head_default']['min_workers'] = int(
                answers['min_workers'])
            self.base_config['available_node_types']['ray_head_default']['max_workers'] = int(
                answers['max_workers'])

        return self.base_config
    
    def verify(self, base_config):
        min_workers = base_config['available_node_types']['ray_head_default']['min_workers']
        max_workers = base_config['available_node_types']['ray_head_default']['max_workers']
        
        if max_workers < min_workers:
            raise Exception(f'specified min workers {min_workers} larger than max workers {max_workers}')

        return base_config



from typing import Any, Dict
from ibm_ray_config.modules.gen2.ssh_key import SshKeyConfig

class RaySshKeyConfig(SshKeyConfig):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)

        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.defaults['key_id'] = self.base_config['available_node_types'][available_node_type]['node_config'].get('key_id')
                break

        self.defaults['ssh_key_filename'] = self.base_config['auth']['ssh_private_key']

    def update_config(self, ssh_key_id, ssh_key_path, ssh_user):        
        self.base_config['auth']['ssh_private_key'] = ssh_key_path
        self.base_config['auth']['ssh_user'] = ssh_user

        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.base_config['available_node_types'][available_node_type]['node_config']['key_id'] = ssh_key_id
        else:
            self.base_config['available_node_types'] = {'ray_head_default': {'node_config': {'key_id': ssh_key_id}}}



from ibm_ray_config.modules.gen2.endpoint import EndpointConfig
from typing import Any, Dict
from ibm_ray_config.modules.utils import get_region_by_endpoint
from ibm_ray_config.modules.config_builder import ConfigBuilder

class RayEndpointConfig(EndpointConfig):

    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)

        base_endpoint = self.base_config['provider'].get('endpoint')
        try:
            # when endpoint was provided directly by user instead of selecting it
            # we just set it
            if base_endpoint:
                ConfigBuilder.ibm_vpc_client.set_service_url(base_endpoint + '/v1')
                self.ibm_vpc_client.set_service_url(base_endpoint + '/v1')

            self.defaults['region'] = get_region_by_endpoint(
                base_endpoint) if base_endpoint else None
        except Exception:
            pass

    def update_config(self, endpoint):
        self.base_config['provider']['endpoint'] = endpoint
        self.base_config['provider']['region'] = ConfigBuilder.region



from typing import Any, Dict

import inquirer
from ibm_ray_config.modules.gen2.vpc import VPCConfig

REQUIRED_RULES = {'outbound_tcp_all': 'selected security group is missing rule permitting outbound TCP access\n', 'outbound_udp_all': 'selected security group is missing rule permitting outbound UDP access\n', 'inbound_tcp_sg': 'selected security group is missing rule permiting inbound tcp traffic inside selected security group\n',
                  'inbound_tcp_22': 'selected security group is missing rule permiting inbound traffic to tcp port 22 required for ssh\n', 'inbound_tcp_6379': 'selected security group is missing rule permiting inbound traffic to tcp port 6379 required for Redis\n', 'inbound_tcp_8265': 'selected security group is missing rule permiting inbound traffic to tcp port 8265 required to access Ray Dashboard\n'}


def validate_security_group(ibm_vpc_client, sec_group_id):
    errors = validate_security_group_rules(ibm_vpc_client, sec_group_id)

    if errors:
        for val in errors.values():
            print(f"\033[91m{val}\033[0m")

        questions = [
            inquirer.List('answer',
                          message='Selected security group is missing required rules, see error above, update with required rules?',
                          choices=['yes', 'no'], default='yes'
                          ), ]

        answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)

        if answers['answer'] == 'yes':
            add_rules_to_security_group(ibm_vpc_client, sec_group_id, errors)
        else:
            exit(1)

        # just in case, validate again the updated/created security group
        errors = validate_security_group_rules(ibm_vpc_client, sec_group_id)
        if not errors:
            return
        else:
            print(f'Something failed during security group rules update/create, please update the required rules manually using ibmcli or web ui and try again')
            exit(1)
    else:
        return


def validate_security_group_rules(ibm_vpc_client, sg_id):
    required_rules = REQUIRED_RULES.copy()

    sg = ibm_vpc_client.get_security_group(sg_id).get_result()

    for rule in sg['rules']:

        # check outbound rules that are not associated with a specific IP address range 
        if rule['direction'] == 'outbound' and rule['remote'] == {'cidr_block': '0.0.0.0/0'}:
            if rule['protocol'] == 'all':
                # outbound is fine!
                required_rules.pop('outbound_tcp_all', None)
                required_rules.pop('outbound_udp_all', None)
            elif rule['protocol'] == 'tcp':
                required_rules.pop('outbound_tcp_all', None)
            elif rule['protocol'] == 'udp':
                required_rules.pop('outbound_udp_all', None)
        
        # Check inbound rules 
        elif rule['direction'] == 'inbound':
            # check rules that are not associated with a specific IP address range
            if rule['remote'] == {'cidr_block': '0.0.0.0/0'}:
                # we interested only in all or tcp protocols
                if rule['protocol'] == 'all':
                    # there a rule permitting all traffic
                    required_rules.pop('inbound_tcp_sg', None)
                    required_rules.pop('inbound_tcp_22', None)
                    required_rules.pop('inbound_tcp_6379', None)
                    required_rules.pop('inbound_tcp_8265', None)

                elif rule['protocol'] == 'tcp':
                    if rule['port_min'] == 1 and rule['port_max'] == 65535:
                        # all ports are open
                        required_rules.pop('inbound_tcp_sg', None)
                        required_rules.pop('inbound_tcp_22', None)
                        required_rules.pop('inbound_tcp_6379', None)
                        required_rules.pop('inbound_tcp_8265', None)
                    else:
                        port_min = rule['port_min']
                        port_max = rule['port_max']
                        if port_min <= 22 and port_max >= 22:
                            required_rules.pop('inbound_tcp_22', None)
                        elif port_min <= 6379 and port_max >= 6379:
                            required_rules.pop('inbound_tcp_6379', None)
                        elif port_min <= 8265 and port_max >= 8265:
                            required_rules.pop('inbound_tcp_8265', None)

            # rule regards private traffic within the VSIs associated with the security group  
            elif rule['remote'].get('id') == sg['id']:
                # validate that inbound traffic inside group available
                if rule['protocol'] == 'all' or rule['protocol'] == 'tcp':
                    required_rules.pop('inbound_tcp_sg', None)

    return required_rules


def add_rules_to_security_group(ibm_vpc_client, sg_id, missing_rules):
    add_rule_msgs = {
        'outbound_tcp_all': f'Add rule to open all outbound TCP ports in selected security group {sg_id}',
        'outbound_udp_all': f'Add rule to open all outbound UDP ports in selected security group {sg_id}',
        'inbound_tcp_sg': f'Add rule to open inbound tcp traffic inside selected security group {sg_id}',
        'inbound_tcp_22': f'Add rule to open inbound tcp port 22 required for SSH in selected security group {sg_id}',
        'inbound_tcp_6379': f'Add rule to open inbound tcp port 6379 required for Redis in selected security group {sg_id}',
        'inbound_tcp_8265': f'Add rule to open inbound tcp port 8265 required to access Ray Dashboard in selected security group {sg_id}'}

    for missing_rule in missing_rules.keys():
        q = [
            inquirer.List('answer',
                          message=add_rule_msgs[missing_rule],
                          choices=['yes', 'no'],
                          default='yes')
        ]

        answers = inquirer.prompt(q, raise_keyboard_interrupt=True)
        if answers['answer'] == 'yes':
            security_group_rule_prototype_model = build_security_group_rule_prototype_model(
                missing_rule, sg_id=sg_id)
            ibm_vpc_client.create_security_group_rule(
                sg_id, security_group_rule_prototype_model).get_result()
        else:
            return False
    return True


def build_security_group_rule_prototype_model(missing_rule, sg_id=None):
    direction, protocol, port = missing_rule.split('_')
    remote = {"cidr_block": "0.0.0.0/0"}

    try:
        port = int(port)
        port_min = port
        port_max = port
    except:
        port_min = 1
        port_max = 65535

        # only valid if security group already exists
        if port == 'sg':
            if not sg_id:
                return None
            remote = {'id': sg_id}

    return {
        'direction': direction,
        'ip_version': 'ipv4',
        'protocol': protocol,
        'remote': remote,
        'port_min': port_min,
        'port_max': port_max
    }


class RayVPCConfig(VPCConfig):

    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)

        self.vpc_name = 'ray-cluster-vpc'
        self.sg_rules = REQUIRED_RULES
        
        if base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.defaults['vpc_id'] = base_config['available_node_types'][available_node_type]['node_config'].get('vpc_id')
                break        
        
    def update_config(self, vpc_obj, zone_obj, subnet_id):
        sec_group_id = vpc_obj['default_security_group']['id']

        validate_security_group(self.ibm_vpc_client, sec_group_id)

        self.base_config['provider']['zone_name'] = zone_obj['name']

        node_config = {
            'vpc_id': vpc_obj['id'],
            'resource_group_id': vpc_obj['resource_group']['id'],
            'security_group_id': sec_group_id,
            'subnet_id': subnet_id
        }

        if self.base_config.get('available_node_types'):
            for available_node_type in self.base_config['available_node_types']:
                self.base_config['available_node_types'][available_node_type]['node_config'].update(
                    node_config)
        else:
            self.base_config['available_node_types'] = {
                'ray_head_default': {'node_config': node_config}}



from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator
from typing import Any, Dict
from ibm_ray_config.modules.utils import color_msg, Color, password_dialog, verify_iam_api_key 


class ApiKeyConfig(ConfigBuilder):

    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)
        self.defaults['api_key'] = self.base_config['ibm']['iam_api_key'] if self.base_config.setdefault('ibm', {}) \
            else None

    @update_decorator
    def run(self, api_key=None, compute_iam_endpoint=None) -> Dict[str, Any]:

        ConfigBuilder.compute_iam_endpoint = compute_iam_endpoint
        
        if not api_key:
            default = self.defaults.get('api_key')
            api_key = password_dialog("Please provide " + color_msg("IBM API KEY", color=Color.CYAN),
                                  default=default,
                                  validate=verify_iam_api_key)['answer']

        ConfigBuilder.iam_api_key = api_key
        return api_key, compute_iam_endpoint

    def update_config(self, iam_api_key, compute_iam_endpoint=None) -> Dict[str, Any]:
        self.base_config['ibm'] = {'iam_api_key': iam_api_key}
        
        if compute_iam_endpoint:
            self.base_config['ibm']['iam_endpoint'] = compute_iam_endpoint
            
        return self.base_config
    
    def verify(self, base_config):
        api_key = base_config['ibm']['iam_api_key']
            
        ConfigBuilder.compute_iam_endpoint = self.base_config['ibm'].get('iam_endpoint')
        
        verify_iam_api_key(None, api_key, iam_endpoint=ConfigBuilder.compute_iam_endpoint)
        ConfigBuilder.iam_api_key = api_key
            
        return base_config




from ibm_ray_config.modules.gen2.ray.api_key import RayApiKeyConfig
from ibm_ray_config.modules.gen2.ray.endpoint import RayEndpointConfig
from ibm_ray_config.modules.gen2.ray.floating_ip import FloatingIpConfig
from ibm_ray_config.modules.gen2.ray.image import RayImageConfig
from ibm_ray_config.modules.gen2.ray.ssh_key import RaySshKeyConfig
from ibm_ray_config.modules.gen2.ray.vpc import RayVPCConfig
from ibm_ray_config.modules.gen2.ray.workers import WorkersConfig
from ibm_ray_config.modules.gen2.ray.profile import RayProfileConfig

MODULES = [RayApiKeyConfig, RayEndpointConfig, RayVPCConfig,
           RaySshKeyConfig, RayImageConfig, FloatingIpConfig, WorkersConfig, RayProfileConfig]

from ibm_ray_config.main import load_base_config

def load_config(backend, iam_api_key, region=None,
                    image_id=None, profile_name='bx2-2x8',
                    key_id=None, ssh_key_filename=None,
                    vpc_id=None, min_workers=0, max_workers=0):
    
    base_config = load_base_config(backend)
    
    base_config['provider']['iam_api_key'] = iam_api_key
    base_config['available_node_types']['ray_head_default']['node_config']['vpc_id'] = vpc_id
    base_config['available_node_types']['ray_head_default']['node_config']['image_id'] = image_id
    base_config['available_node_types']['ray_head_default']['node_config']['instance_profile_name'] = profile_name
    base_config['available_node_types']['ray_head_default']['node_config']['key_id'] = key_id
    base_config['auth']['ssh_private_key'] = ssh_key_filename
    
    base_config['provider']['region'] = region
    base_config['provider']['endpoint'] = f'https://{region}.iaas.cloud.ibm.com'

    base_config['max_workers'] = max_workers
    base_config['available_node_types']['ray_head_default']['min_workers'] = min_workers
    base_config['available_node_types']['ray_head_default']['max_workers'] = max_workers
    
    return base_config

def parse_config(config):
    res = {'iam_api_key': config['provider']['iam_api_key']}    
    
    for available_node_type in config['available_node_types']:
        res['vpc_id'] = config['available_node_types'][available_node_type]['node_config']['vpc_id']
        res['key_id'] = config['available_node_types'][available_node_type]['node_config']['key_id']     
        res['subnet_id'] = config['available_node_types'][available_node_type]['node_config']['subnet_id']
    
    res['endpoint'] = config['provider']['endpoint']

    if 'iam_endpoint' in config['provider']:
        res['iam_endpoint'] = config['provider']['iam_endpoint']
        
    return res
    